# Copyright 2017 Inspur Corp.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
"""Tests for the Inspur InStorage volume driver."""

from unittest import mock

from eventlet import greenthread
from oslo_utils import importutils
import six

from cinder import context
from cinder import exception
from cinder.tests.unit import test
from cinder.tests.unit import utils as testutils
from cinder.tests.unit.volume.drivers.inspur.instorage import fakes
from cinder.volume import configuration as conf
from cinder.volume.drivers.inspur.instorage import instorage_iscsi
from cinder.volume import volume_types


class InStorageMCSISCSIDriverTestCase(test.TestCase):
    def setUp(self):
        super(InStorageMCSISCSIDriverTestCase, self).setUp()
        self.mock_object(greenthread, 'sleep')
        self.iscsi_driver = fakes.FakeInStorageMCSISCSIDriver(
            configuration=conf.Configuration(None))
        self._def_flags = {'san_ip': 'hostname',
                           'san_login': 'user',
                           'san_password': 'pass',
                           'instorage_mcs_volpool_name': ['openstack'],
                           'instorage_mcs_localcopy_timeout': 20,
                           'instorage_mcs_localcopy_rate': 49,
                           'instorage_mcs_allow_tenant_qos': True}
        wwpns = ['1234567890123456', '6543210987654321']
        initiator = 'test.initiator.%s' % 123456
        self._connector = {'ip': '1.234.56.78',
                           'host': 'instorage-mcs-test',
                           'wwpns': wwpns,
                           'initiator': initiator}
        self.sim = fakes.FakeInStorage(['openstack'])

        self.iscsi_driver.set_fake_storage(self.sim)
        self.ctxt = context.get_admin_context()

        self._reset_flags()
        self.ctxt = context.get_admin_context()
        db_driver = self.iscsi_driver.configuration.db_driver
        self.db = importutils.import_module(db_driver)
        self.iscsi_driver.db = self.db
        self.iscsi_driver.do_setup(None)
        self.iscsi_driver.check_for_setup_error()
        self.iscsi_driver._assistant.check_lcmapping_interval = 0

    def _set_flag(self, flag, value):
        group = self.iscsi_driver.configuration.config_group
        self.iscsi_driver.configuration.set_override(flag, value, group)

    def _reset_flags(self):
        self.iscsi_driver.configuration.local_conf.reset()
        for k, v in self._def_flags.items():
            self._set_flag(k, v)

    def _create_volume(self, **kwargs):
        pool = fakes.get_test_pool()
        prop = {'host': 'openstack@mcs#%s' % pool,
                'size': 1,
                'volume_type_id': self.vt['id']}
        for p in prop.keys():
            if p not in kwargs:
                kwargs[p] = prop[p]
        vol = testutils.create_volume(self.ctxt, **kwargs)
        self.iscsi_driver.create_volume(vol)
        return vol

    def _delete_volume(self, volume):
        self.iscsi_driver.delete_volume(volume)
        self.db.volume_destroy(self.ctxt, volume['id'])

    def _generate_vol_info(self, vol_name, vol_id):
        pool = fakes.get_test_pool()
        prop = {'mdisk_grp_name': pool}
        if vol_name:
            prop.update(volume_name=vol_name,
                        volume_id=vol_id,
                        volume_size=10)
        else:
            prop.update(size=10,
                        volume_type_id=None,
                        mdisk_grp_name=pool,
                        host='openstack@mcs#%s' % pool)
        vol = testutils.create_volume(self.ctxt, **prop)
        return vol

    def _assert_vol_exists(self, name, exists):
        is_vol_defined = self.iscsi_driver._assistant.is_vdisk_defined(name)
        self.assertEqual(exists, is_vol_defined)

    def test_instorage_mcs_iscsi_validate_connector(self):
        conn_neither = {'host': 'host'}
        conn_iscsi = {'host': 'host', 'initiator': 'foo'}
        conn_fc = {'host': 'host', 'wwpns': 'bar'}
        conn_both = {'host': 'host', 'initiator': 'foo', 'wwpns': 'bar'}

        self.iscsi_driver._state['enabled_protocols'] = set(['iSCSI'])
        self.iscsi_driver.validate_connector(conn_iscsi)
        self.iscsi_driver.validate_connector(conn_both)
        self.assertRaises(exception.InvalidConnectorException,
                          self.iscsi_driver.validate_connector, conn_fc)
        self.assertRaises(exception.InvalidConnectorException,
                          self.iscsi_driver.validate_connector, conn_neither)

        self.iscsi_driver._state['enabled_protocols'] = set(['iSCSI', 'FC'])
        self.iscsi_driver.validate_connector(conn_iscsi)
        self.iscsi_driver.validate_connector(conn_both)
        self.assertRaises(exception.InvalidConnectorException,
                          self.iscsi_driver.validate_connector, conn_neither)

    def test_instorage_terminate_iscsi_connection(self):
        # create a iSCSI volume
        volume_iSCSI = self._create_volume()
        extra_spec = {'capabilities:storage_protocol': '<in> iSCSI'}
        vol_type_iSCSI = volume_types.create(self.ctxt, 'iSCSI', extra_spec)
        volume_iSCSI['volume_type_id'] = vol_type_iSCSI['id']

        connector = {'host': 'instorage-mcs-host',
                     'wwnns': ['20000090fa17311e', '20000090fa17311f'],
                     'wwpns': ['ff00000000000000', 'ff00000000000001'],
                     'initiator': 'iqn.1993-08.org.debian:01:eac5ccc1aaa'}

        self.iscsi_driver.initialize_connection(volume_iSCSI, connector)
        self.iscsi_driver.terminate_connection(volume_iSCSI, connector)

    @mock.patch.object(instorage_iscsi.InStorageMCSISCSIDriver,
                       '_do_terminate_connection')
    def test_instorage_initialize_iscsi_connection_failure(self, term_conn):
        # create a iSCSI volume
        volume_iSCSI = self._create_volume()
        extra_spec = {'capabilities:storage_protocol': '<in> iSCSI'}
        vol_type_iSCSI = volume_types.create(self.ctxt, 'iSCSI', extra_spec)
        volume_iSCSI['volume_type_id'] = vol_type_iSCSI['id']

        connector = {'host': 'instorage-mcs-host',
                     'wwnns': ['20000090fa17311e', '20000090fa17311f'],
                     'wwpns': ['ff00000000000000', 'ff00000000000001'],
                     'initiator': 'iqn.1993-08.org.debian:01:eac5ccc1aaa'}

        self.iscsi_driver._state['storage_nodes'] = {}
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.iscsi_driver.initialize_connection,
                          volume_iSCSI, connector)
        term_conn.assert_called_once_with(volume_iSCSI, connector)

    def test_instorage_initialize_iscsi_connection_multihost(self):
        connector_a = {'host': 'instorage-mcs-host-a',
                       'wwnns': ['20000090fa17311e', '20000090fa17311f'],
                       'wwpns': ['ff00000000000000', 'ff00000000000001'],
                       'initiator': 'iqn.1993-08.org.debian:01:eac5ccc1aaa'}
        # host-volume map return value
        exp_path_a = {'driver_volume_type': 'iscsi',
                      'data': {'target_discovered': False,
                               'target_iqn':
                                   'iqn.1982-01.com.inspur:1234.sim.node1',
                               'target_portal': '1.234.56.78:3260',
                               'target_lun': 0,
                               'auth_method': 'CHAP',
                               'discovery_auth_method': 'CHAP'}}

        connector_b = {'host': 'instorage-mcs-host-b',
                       'wwnns': ['30000090fa17311e', '30000090fa17311f'],
                       'wwpns': ['ff00000000000002', 'ff00000000000003'],
                       'initiator': 'iqn.1993-08.org.debian:01:eac5ccc1aab'}
        # host-volume map return value
        exp_path_b = {'driver_volume_type': 'iscsi',
                      'data': {'target_discovered': False,
                               'target_iqn':
                                   'iqn.1982-01.com.inspur:1234.sim.node1',
                               'target_portal': '1.234.56.78:3260',
                               'target_lun': 1,
                               'auth_method': 'CHAP',
                               'discovery_auth_method': 'CHAP'}}

        volume_iSCSI = self._create_volume()
        extra_spec = {'capabilities:storage_protocol': '<in> iSCSI'}
        vol_type_iSCSI = volume_types.create(self.ctxt, 'iSCSI', extra_spec)
        volume_iSCSI['volume_type_id'] = vol_type_iSCSI['id']

        # Make sure that the volumes have been created
        self._assert_vol_exists(volume_iSCSI['name'], True)

        # check that the hosts not exist
        ret = self.iscsi_driver._assistant.get_host_from_connector(
            connector_a)
        self.assertIsNone(ret)
        ret = self.iscsi_driver._assistant.get_host_from_connector(
            connector_b)
        self.assertIsNone(ret)

        # Initialize connection to map volume to host a
        ret = self.iscsi_driver.initialize_connection(
            volume_iSCSI, connector_a)
        self.assertEqual(exp_path_a['driver_volume_type'],
                         ret['driver_volume_type'])

        # check host-volume map return value
        for k, v in exp_path_a['data'].items():
            self.assertEqual(v, ret['data'][k])

        ret = self.iscsi_driver._assistant.get_host_from_connector(
            connector_a)
        self.assertIsNotNone(ret)

        # Initialize connection to map volume to host b
        ret = self.iscsi_driver.initialize_connection(
            volume_iSCSI, connector_b)
        self.assertEqual(exp_path_b['driver_volume_type'],
                         ret['driver_volume_type'])

        # check the return value
        for k, v in exp_path_b['data'].items():
            self.assertEqual(v, ret['data'][k])

        ret = self.iscsi_driver._assistant.get_host_from_connector(
            connector_b)
        self.assertIsNotNone(ret)

    def test_instorage_initialize_iscsi_connection_single_path(self):
        # Test the return value for _get_iscsi_properties

        connector = {'host': 'instorage-mcs-host',
                     'wwnns': ['20000090fa17311e', '20000090fa17311f'],
                     'wwpns': ['ff00000000000000', 'ff00000000000001'],
                     'initiator': 'iqn.1993-08.org.debian:01:eac5ccc1aaa'}
        # Expected single path host-volume map return value
        exp_s_path = {'driver_volume_type': 'iscsi',
                      'data': {'target_discovered': False,
                               'target_iqn':
                                   'iqn.1982-01.com.inspur:1234.sim.node1',
                               'target_portal': '1.234.56.78:3260',
                               'target_lun': 0,
                               'auth_method': 'CHAP',
                               'discovery_auth_method': 'CHAP'}}

        volume_iSCSI = self._create_volume()
        extra_spec = {'capabilities:storage_protocol': '<in> iSCSI'}
        vol_type_iSCSI = volume_types.create(self.ctxt, 'iSCSI', extra_spec)
        volume_iSCSI['volume_type_id'] = vol_type_iSCSI['id']

        # Make sure that the volumes have been created
        self._assert_vol_exists(volume_iSCSI['name'], True)

        # Check case where no hosts exist
        ret = self.iscsi_driver._assistant.get_host_from_connector(
            connector)
        self.assertIsNone(ret)

        # Initialize connection to map volume to a host
        ret = self.iscsi_driver.initialize_connection(
            volume_iSCSI, connector)
        self.assertEqual(exp_s_path['driver_volume_type'],
                         ret['driver_volume_type'])

        # Check the single path host-volume map return value
        for k, v in exp_s_path['data'].items():
            self.assertEqual(v, ret['data'][k])

        ret = self.iscsi_driver._assistant.get_host_from_connector(
            connector)
        self.assertIsNotNone(ret)

    def test_instorage_initialize_iscsi_connection_multipath(self):
        # Test the return value for _get_iscsi_properties

        connector = {'host': 'instorage-mcs-host',
                     'wwnns': ['20000090fa17311e', '20000090fa17311f'],
                     'wwpns': ['ff00000000000000', 'ff00000000000001'],
                     'initiator': 'iqn.1993-08.org.debian:01:eac5ccc1aaa',
                     'multipath': True}

        # Expected multipath host-volume map return value
        exp_m_path = {'driver_volume_type': 'iscsi',
                      'data': {'target_discovered': False,
                               'target_iqn':
                                   'iqn.1982-01.com.inspur:1234.sim.node1',
                               'target_portal': '1.234.56.78:3260',
                               'target_lun': 0,
                               'target_iqns': [
                                   'iqn.1982-01.com.inspur:1234.sim.node1',
                                   'iqn.1982-01.com.inspur:1234.sim.node1',
                                   'iqn.1982-01.com.inspur:1234.sim.node2'],
                               'target_portals':
                                   ['1.234.56.78:3260',
                                    '1.234.56.80:3260',
                                    '1.234.56.79:3260'],
                               'target_luns': [0, 0, 0],
                               'auth_method': 'CHAP',
                               'discovery_auth_method': 'CHAP'}}

        volume_iSCSI = self._create_volume()
        extra_spec = {'capabilities:storage_protocol': '<in> iSCSI'}
        vol_type_iSCSI = volume_types.create(self.ctxt, 'iSCSI', extra_spec)
        volume_iSCSI['volume_type_id'] = vol_type_iSCSI['id']

        # Check case where no hosts exist
        ret = self.iscsi_driver._assistant.get_host_from_connector(
            connector)
        self.assertIsNone(ret)

        # Initialize connection to map volume to a host
        ret = self.iscsi_driver.initialize_connection(
            volume_iSCSI, connector)
        self.assertEqual(exp_m_path['driver_volume_type'],
                         ret['driver_volume_type'])

        # Check the multipath host-volume map return value
        for k, v in exp_m_path['data'].items():
            if k in ('target_iqns', 'target_portals'):
                # These are randomly ordered lists
                six.assertCountEqual(self, v, ret['data'][k])
            else:
                self.assertEqual(v, ret['data'][k])

        ret = self.iscsi_driver._assistant.get_host_from_connector(
            connector)
        self.assertIsNotNone(ret)

    def test_instorage_mcs_iscsi_host_maps(self):
        # Create two volumes to be used in mappings

        ctxt = context.get_admin_context()
        volume1 = self._generate_vol_info(None, None)
        self.iscsi_driver.create_volume(volume1)
        volume2 = self._generate_vol_info(None, None)
        self.iscsi_driver.create_volume(volume2)

        # Create volume types that we created
        types = {}
        for protocol in ['iSCSI']:
            opts = {'storage_protocol': '<in> ' + protocol}
            types[protocol] = volume_types.create(ctxt, protocol, opts)

        expected = {'iSCSI': {'driver_volume_type': 'iscsi',
                              'data': {'target_discovered': False,
                                       'target_iqn':
                                       'iqn.1982-01.com.inspur:1234.sim.node1',
                                       'target_portal': '1.234.56.78:3260',
                                       'target_lun': 0,
                                       'auth_method': 'CHAP',
                                       'discovery_auth_method': 'CHAP'}}}

        volume1['volume_type_id'] = types[protocol]['id']
        volume2['volume_type_id'] = types[protocol]['id']

        # Check case where no hosts exist
        ret = self.iscsi_driver._assistant.get_host_from_connector(
            self._connector)
        self.assertIsNone(ret)

        # Make sure that the volumes have been created
        self._assert_vol_exists(volume1['name'], True)
        self._assert_vol_exists(volume2['name'], True)

        # Initialize connection from the first volume to a host
        ret = self.iscsi_driver.initialize_connection(
            volume1, self._connector)
        self.assertEqual(expected[protocol]['driver_volume_type'],
                         ret['driver_volume_type'])
        for k, v in expected[protocol]['data'].items():
            self.assertEqual(v, ret['data'][k])

        # Initialize again, should notice it and do nothing
        ret = self.iscsi_driver.initialize_connection(
            volume1, self._connector)
        self.assertEqual(expected[protocol]['driver_volume_type'],
                         ret['driver_volume_type'])
        for k, v in expected[protocol]['data'].items():
            self.assertEqual(v, ret['data'][k])

        # Try to delete the 1st volume (should fail because it is mapped)
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.iscsi_driver.delete_volume,
                          volume1)

        ret = self.iscsi_driver.terminate_connection(volume1,
                                                     self._connector)
        ret = self.iscsi_driver._assistant.get_host_from_connector(
            self._connector)
        self.assertIsNone(ret)

        # Check cases with no auth set for host
        for auth_enabled in [True, False]:
            for host_exists in ['yes-auth', 'yes-noauth', 'no']:
                self._set_flag('instorage_mcs_iscsi_chap_enabled',
                               auth_enabled)
                case = 'en' + six.text_type(
                    auth_enabled) + 'ex' + six.text_type(host_exists)
                conn_na = {'initiator': 'test:init:%s' % 56789,
                           'ip': '11.11.11.11',
                           'host': 'host-%s' % case}
                if host_exists.startswith('yes'):
                    self.sim._add_host_to_list(conn_na)
                    if host_exists == 'yes-auth':
                        kwargs = {'chapsecret': 'foo',
                                  'obj': conn_na['host']}
                        self.sim._cmd_chhost(**kwargs)
                volume1['volume_type_id'] = types['iSCSI']['id']

                init_ret = self.iscsi_driver.initialize_connection(volume1,
                                                                   conn_na)
                host_name = self.sim._host_in_list(conn_na['host'])
                chap_ret = (
                    self.iscsi_driver._assistant.get_chap_secret_for_host(
                        host_name))
                if auth_enabled or host_exists == 'yes-auth':
                    self.assertIn('auth_password', init_ret['data'])
                    self.assertIsNotNone(chap_ret)
                else:
                    self.assertNotIn('auth_password', init_ret['data'])
                    self.assertIsNone(chap_ret)
                self.iscsi_driver.terminate_connection(volume1, conn_na)
        self._set_flag('instorage_mcs_iscsi_chap_enabled', True)

        # Test no preferred node
        self.sim.error_injection('lsvdisk', 'no_pref_node')
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.iscsi_driver.initialize_connection,
                          volume1, self._connector)

        # Initialize connection from the second volume to the host with no
        # preferred node set if in simulation mode, otherwise, just
        # another initialize connection.
        self.sim.error_injection('lsvdisk', 'blank_pref_node')
        self.iscsi_driver.initialize_connection(volume2, self._connector)

        # Try to remove connection from host that doesn't exist (should fail)
        conn_no_exist = self._connector.copy()
        conn_no_exist['initiator'] = 'i_dont_exist'
        conn_no_exist['wwpns'] = ['0000000000000000']
        self.assertRaises(exception.VolumeDriverException,
                          self.iscsi_driver.terminate_connection,
                          volume1,
                          conn_no_exist)

        # Try to remove connection from volume that isn't mapped (should print
        # message but NOT fail)
        unmapped_vol = self._generate_vol_info(None, None)
        self.iscsi_driver.create_volume(unmapped_vol)
        self.iscsi_driver.terminate_connection(unmapped_vol, self._connector)
        self.iscsi_driver.delete_volume(unmapped_vol)

        # Remove the mapping from the 1st volume and delete it
        self.iscsi_driver.terminate_connection(volume1, self._connector)
        self.iscsi_driver.delete_volume(volume1)
        self._assert_vol_exists(volume1['name'], False)

        # Make sure our host still exists
        host_name = self.iscsi_driver._assistant.get_host_from_connector(
            self._connector)
        self.assertIsNotNone(host_name)

        # Remove the mapping from the 2nd volume. The host should
        # be automatically removed because there are no more mappings.
        self.iscsi_driver.terminate_connection(volume2, self._connector)

        # Check if we successfully terminate connections when the host is not
        # specified
        fake_conn = {'ip': '127.0.0.1', 'initiator': 'iqn.fake'}
        self.iscsi_driver.initialize_connection(volume2, self._connector)
        host_name = self.iscsi_driver._assistant.get_host_from_connector(
            self._connector)
        self.assertIsNotNone(host_name)
        self.iscsi_driver.terminate_connection(volume2, fake_conn)
        host_name = self.iscsi_driver._assistant.get_host_from_connector(
            self._connector)
        self.assertIsNone(host_name)
        self.iscsi_driver.delete_volume(volume2)
        self._assert_vol_exists(volume2['name'], False)

        # Delete volume types that we created
        for protocol in ['iSCSI']:
            volume_types.destroy(ctxt, types[protocol]['id'])

        # Check if our host still exists (it should not)
        ret = (self.iscsi_driver._assistant.get_host_from_connector(
            self._connector))
        self.assertIsNone(ret)

    def test_add_vdisk_copy_iscsi(self):
        # Ensure only iSCSI is available
        self.iscsi_driver._state['enabled_protocols'] = set(['iSCSI'])
        volume = self._generate_vol_info(None, None)
        self.iscsi_driver.create_volume(volume)
        self.iscsi_driver.add_vdisk_copy(volume['name'], 'fake-pool', None)



#    Copyright 2015 SimpliVity Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
from unittest import mock

import ddt
from oslo_utils import timeutils
import pytz
import six

from cinder.db.sqlalchemy import models
from cinder import exception
from cinder import objects
from cinder.objects import fields
from cinder.tests.unit import fake_constants as fake
from cinder.tests.unit import fake_snapshot
from cinder.tests.unit import fake_volume
from cinder.tests.unit import objects as test_objects


fake_db_snapshot = fake_snapshot.fake_db_snapshot(
    cgsnapshot_id=fake.CGSNAPSHOT_ID)
del fake_db_snapshot['metadata']
del fake_db_snapshot['volume']


# NOTE(andrey-mp): make Snapshot object here to check object algorithms
fake_snapshot_obj = {
    'id': fake.SNAPSHOT_ID,
    'volume_id': fake.VOLUME_ID,
    'status': fields.SnapshotStatus.CREATING,
    'progress': '0%',
    'volume_size': 1,
    'display_name': 'fake_name',
    'display_description': 'fake_description',
    'metadata': {},
}


@ddt.ddt
class TestSnapshot(test_objects.BaseObjectsTestCase):

    @mock.patch('cinder.db.get_by_id', return_value=fake_db_snapshot)
    def test_get_by_id(self, snapshot_get):
        snapshot = objects.Snapshot.get_by_id(self.context, 1)
        self._compare(self, fake_snapshot_obj, snapshot)
        snapshot_get.assert_called_once_with(self.context, models.Snapshot, 1)

    @mock.patch('cinder.db.sqlalchemy.api.model_query')
    def test_get_by_id_no_existing_id(self, model_query):
        query = model_query().options().options().filter_by().first
        query.return_value = None
        self.assertRaises(exception.SnapshotNotFound,
                          objects.Snapshot.get_by_id, self.context, 123)

    def test_reset_changes(self):
        snapshot = objects.Snapshot()
        snapshot.metadata = {'key1': 'value1'}
        self.assertEqual({}, snapshot._orig_metadata)
        snapshot.obj_reset_changes(['metadata'])
        self.assertEqual({'key1': 'value1'}, snapshot._orig_metadata)

    @mock.patch('cinder.db.snapshot_create', return_value=fake_db_snapshot)
    def test_create(self, snapshot_create):
        snapshot = objects.Snapshot(context=self.context)
        snapshot.create()
        self.assertEqual(fake_snapshot_obj['id'], snapshot.id)
        self.assertEqual(fake_snapshot_obj['volume_id'], snapshot.volume_id)

    @mock.patch('cinder.db.snapshot_create')
    def test_create_with_provider_id(self, snapshot_create):
        snapshot_create.return_value = copy.deepcopy(fake_db_snapshot)
        snapshot_create.return_value['provider_id'] = fake.PROVIDER_ID

        snapshot = objects.Snapshot(context=self.context)
        snapshot.create()
        self.assertEqual(fake.PROVIDER_ID, snapshot.provider_id)

    @mock.patch('cinder.db.snapshot_update')
    def test_save(self, snapshot_update):
        snapshot = objects.Snapshot._from_db_object(
            self.context, objects.Snapshot(), fake_db_snapshot)
        snapshot.display_name = 'foobar'
        snapshot.save()
        snapshot_update.assert_called_once_with(self.context, snapshot.id,
                                                {'display_name': 'foobar'})

    @mock.patch('cinder.db.snapshot_metadata_update',
                return_value={'key1': 'value1'})
    @mock.patch('cinder.db.snapshot_update')
    def test_save_with_metadata(self, snapshot_update,
                                snapshot_metadata_update):
        snapshot = objects.Snapshot._from_db_object(
            self.context, objects.Snapshot(), fake_db_snapshot)
        snapshot.display_name = 'foobar'
        snapshot.metadata = {'key1': 'value1'}
        self.assertEqual({'display_name': 'foobar',
                          'metadata': {'key1': 'value1'}},
                         snapshot.obj_get_changes())
        snapshot.save()
        snapshot_update.assert_called_once_with(self.context, snapshot.id,
                                                {'display_name': 'foobar'})
        snapshot_metadata_update.assert_called_once_with(self.context,
                                                         fake.SNAPSHOT_ID,
                                                         {'key1': 'value1'},
                                                         True)

    @mock.patch('oslo_utils.timeutils.utcnow', return_value=timeutils.utcnow())
    @mock.patch('cinder.db.sqlalchemy.api.snapshot_destroy')
    def test_destroy(self, snapshot_destroy, utcnow_mock):
        snapshot_destroy.return_value = {
            'status': 'deleted',
            'deleted': True,
            'deleted_at': utcnow_mock.return_value}
        snapshot = objects.Snapshot(context=self.context,
                                    id=fake.SNAPSHOT_ID)
        snapshot.destroy()
        self.assertTrue(snapshot_destroy.called)
        admin_context = snapshot_destroy.call_args[0][0]
        self.assertTrue(admin_context.is_admin)
        self.assertTrue(snapshot.deleted)
        self.assertEqual(fields.SnapshotStatus.DELETED, snapshot.status)
        self.assertEqual(utcnow_mock.return_value.replace(tzinfo=pytz.UTC),
                         snapshot.deleted_at)

    @mock.patch('cinder.db.snapshot_metadata_delete')
    def test_delete_metadata_key(self, snapshot_metadata_delete):
        snapshot = objects.Snapshot(self.context, id=fake.SNAPSHOT_ID)
        snapshot.metadata = {'key1': 'value1', 'key2': 'value2'}
        self.assertEqual({}, snapshot._orig_metadata)
        snapshot.delete_metadata_key(self.context, 'key2')
        self.assertEqual({'key1': 'value1'}, snapshot.metadata)
        snapshot_metadata_delete.assert_called_once_with(self.context,
                                                         fake.SNAPSHOT_ID,
                                                         'key2')

    def test_obj_fields(self):
        volume = objects.Volume(context=self.context, id=fake.VOLUME_ID,
                                _name_id=fake.VOLUME_NAME_ID)
        snapshot = objects.Snapshot(context=self.context, id=fake.VOLUME_ID,
                                    volume=volume)
        self.assertEqual(['name', 'volume_name'], snapshot.obj_extra_fields)
        self.assertEqual('snapshot-%s' % fake.VOLUME_ID, snapshot.name)
        self.assertEqual('volume-%s' % fake.VOLUME_NAME_ID,
                         snapshot.volume_name)

    @mock.patch('cinder.objects.volume.Volume.get_by_id')
    @mock.patch('cinder.objects.cgsnapshot.CGSnapshot.get_by_id')
    def test_obj_load_attr(self, cgsnapshot_get_by_id, volume_get_by_id):
        snapshot = objects.Snapshot._from_db_object(
            self.context, objects.Snapshot(), fake_db_snapshot)
        # Test volume lazy-loaded field
        volume = objects.Volume(context=self.context, id=fake.VOLUME_ID)
        volume_get_by_id.return_value = volume
        self.assertEqual(volume, snapshot.volume)
        volume_get_by_id.assert_called_once_with(self.context,
                                                 snapshot.volume_id)
        self.assertEqual(snapshot.metadata, {})
        # Test cgsnapshot lazy-loaded field
        cgsnapshot = objects.CGSnapshot(context=self.context,
                                        id=fake.CGSNAPSHOT_ID)
        cgsnapshot_get_by_id.return_value = cgsnapshot
        self.assertEqual(cgsnapshot, snapshot.cgsnapshot)
        cgsnapshot_get_by_id.assert_called_once_with(self.context,
                                                     snapshot.cgsnapshot_id)

    @mock.patch('cinder.objects.cgsnapshot.CGSnapshot.get_by_id')
    def test_obj_load_attr_cgroup_not_exist(self, cgsnapshot_get_by_id):
        fake_non_cg_db_snapshot = fake_snapshot.fake_db_snapshot(
            cgsnapshot_id=None)
        snapshot = objects.Snapshot._from_db_object(
            self.context, objects.Snapshot(), fake_non_cg_db_snapshot)
        self.assertIsNone(snapshot.cgsnapshot)
        cgsnapshot_get_by_id.assert_not_called()

    @mock.patch('cinder.objects.group_snapshot.GroupSnapshot.get_by_id')
    def test_obj_load_attr_group_not_exist(self, group_snapshot_get_by_id):
        fake_non_cg_db_snapshot = fake_snapshot.fake_db_snapshot(
            group_snapshot_id=None)
        snapshot = objects.Snapshot._from_db_object(
            self.context, objects.Snapshot(), fake_non_cg_db_snapshot)
        self.assertIsNone(snapshot.group_snapshot)
        group_snapshot_get_by_id.assert_not_called()

    @mock.patch('cinder.db.snapshot_data_get_for_project')
    def test_snapshot_data_get_for_project(self, snapshot_data_get):
        snapshot = objects.Snapshot._from_db_object(
            self.context, objects.Snapshot(), fake_db_snapshot)
        volume_type_id = mock.sentinel.volume_type_id
        snapshot.snapshot_data_get_for_project(self.context,
                                               self.project_id,
                                               volume_type_id)
        snapshot_data_get.assert_called_once_with(self.context,
                                                  self.project_id,
                                                  volume_type_id,
                                                  host=None)

    @mock.patch('cinder.db.sqlalchemy.api.snapshot_get')
    def test_refresh(self, snapshot_get):
        db_snapshot1 = fake_snapshot.fake_db_snapshot()
        db_snapshot2 = db_snapshot1.copy()
        db_snapshot2['display_name'] = 'foobar'

        # On the second snapshot_get, return the snapshot with an updated
        # display_name
        snapshot_get.side_effect = [db_snapshot1, db_snapshot2]
        snapshot = objects.Snapshot.get_by_id(self.context, fake.SNAPSHOT_ID)
        self._compare(self, db_snapshot1, snapshot)

        # display_name was updated, so a snapshot refresh should have a new
        # value for that field
        snapshot.refresh()
        self._compare(self, db_snapshot2, snapshot)
        if six.PY3:
            call_bool = mock.call.__bool__()
        else:
            call_bool = mock.call.__nonzero__()
        snapshot_get.assert_has_calls([
            mock.call(self.context,
                      fake.SNAPSHOT_ID),
            call_bool,
            mock.call(self.context,
                      fake.SNAPSHOT_ID)])

    @ddt.data('1.1', '1.3')
    def test_obj_make_compatible_1_3(self, version):
        snapshot = objects.Snapshot(context=self.context)
        snapshot.status = fields.SnapshotStatus.UNMANAGING
        primitive = snapshot.obj_to_primitive(version)
        snapshot = objects.Snapshot.obj_from_primitive(primitive)
        if version == '1.3':
            status = fields.SnapshotStatus.UNMANAGING
        else:
            status = fields.SnapshotStatus.DELETING
        self.assertEqual(status, snapshot.status)

    @ddt.data('1.3', '1.4')
    def test_obj_make_compatible_1_4(self, version):
        snapshot = objects.Snapshot(context=self.context)
        snapshot.status = fields.SnapshotStatus.BACKING_UP
        primitive = snapshot.obj_to_primitive(version)
        snapshot = objects.Snapshot.obj_from_primitive(primitive)
        if version == '1.4':
            status = fields.SnapshotStatus.BACKING_UP
        else:
            status = fields.SnapshotStatus.AVAILABLE
        self.assertEqual(status, snapshot.status)


class TestSnapshotList(test_objects.BaseObjectsTestCase):
    @mock.patch('cinder.objects.volume.Volume.get_by_id')
    @mock.patch('cinder.db.snapshot_get_all', return_value=[fake_db_snapshot])
    def test_get_all(self, snapshot_get_all, volume_get_by_id):
        fake_volume_obj = fake_volume.fake_volume_obj(self.context)
        volume_get_by_id.return_value = fake_volume_obj

        search_opts = mock.sentinel.search_opts
        snapshots = objects.SnapshotList.get_all(
            self.context, search_opts)
        self.assertEqual(1, len(snapshots))
        TestSnapshot._compare(self, fake_snapshot_obj, snapshots[0])
        snapshot_get_all.assert_called_once_with(self.context, search_opts,
                                                 None, None, None, None, None)

    @mock.patch('cinder.objects.Volume.get_by_id')
    @mock.patch('cinder.db.snapshot_get_all_by_host',
                return_value=[fake_db_snapshot])
    def test_get_by_host(self, get_by_host, volume_get_by_id):
        fake_volume_obj = fake_volume.fake_volume_obj(self.context)
        volume_get_by_id.return_value = fake_volume_obj

        snapshots = objects.SnapshotList.get_by_host(
            self.context, 'fake-host')
        self.assertEqual(1, len(snapshots))
        TestSnapshot._compare(self, fake_snapshot_obj, snapshots[0])

    @mock.patch('cinder.objects.volume.Volume.get_by_id')
    @mock.patch('cinder.db.snapshot_get_all_by_project',
                return_value=[fake_db_snapshot])
    def test_get_all_by_project(self, get_all_by_project, volume_get_by_id):
        fake_volume_obj = fake_volume.fake_volume_obj(self.context)
        volume_get_by_id.return_value = fake_volume_obj

        search_opts = mock.sentinel.search_opts
        snapshots = objects.SnapshotList.get_all_by_project(
            self.context, self.project_id, search_opts)
        self.assertEqual(1, len(snapshots))
        TestSnapshot._compare(self, fake_snapshot_obj, snapshots[0])
        get_all_by_project.assert_called_once_with(self.context,
                                                   self.project_id,
                                                   search_opts, None, None,
                                                   None, None, None)

    @mock.patch('cinder.objects.volume.Volume.get_by_id')
    @mock.patch('cinder.db.snapshot_get_all_for_volume',
                return_value=[fake_db_snapshot])
    def test_get_all_for_volume(self, get_all_for_volume, volume_get_by_id):
        fake_volume_obj = fake_volume.fake_volume_obj(self.context)
        volume_get_by_id.return_value = fake_volume_obj

        snapshots = objects.SnapshotList.get_all_for_volume(
            self.context, fake_volume_obj.id)
        self.assertEqual(1, len(snapshots))
        TestSnapshot._compare(self, fake_snapshot_obj, snapshots[0])

    @mock.patch('cinder.objects.volume.Volume.get_by_id')
    @mock.patch('cinder.db.snapshot_get_all_active_by_window',
                return_value=[fake_db_snapshot])
    def test_get_all_active_by_window(self, get_all_active_by_window,
                                      volume_get_by_id):
        fake_volume_obj = fake_volume.fake_volume_obj(self.context)
        volume_get_by_id.return_value = fake_volume_obj

        snapshots = objects.SnapshotList.get_all_active_by_window(
            self.context, mock.sentinel.begin, mock.sentinel.end)
        self.assertEqual(1, len(snapshots))
        TestSnapshot._compare(self, fake_snapshot_obj, snapshots[0])

    @mock.patch('cinder.objects.volume.Volume.get_by_id')
    @mock.patch('cinder.db.snapshot_get_all_for_cgsnapshot',
                return_value=[fake_db_snapshot])
    def test_get_all_for_cgsnapshot(self, get_all_for_cgsnapshot,
                                    volume_get_by_id):
        fake_volume_obj = fake_volume.fake_volume_obj(self.context)
        volume_get_by_id.return_value = fake_volume_obj

        snapshots = objects.SnapshotList.get_all_for_cgsnapshot(
            self.context, mock.sentinel.cgsnapshot_id)
        self.assertEqual(1, len(snapshots))
        TestSnapshot._compare(self, fake_snapshot_obj, snapshots[0])

    @mock.patch('cinder.objects.volume.Volume.get_by_id')
    @mock.patch('cinder.db.snapshot_get_all')
    def test_get_all_without_metadata(self, snapshot_get_all,
                                      volume_get_by_id):
        fake_volume_obj = fake_volume.fake_volume_obj(self.context)
        volume_get_by_id.return_value = fake_volume_obj

        snapshot = copy.deepcopy(fake_db_snapshot)
        del snapshot['snapshot_metadata']
        snapshot_get_all.return_value = [snapshot]

        search_opts = mock.sentinel.search_opts
        self.assertRaises(exception.MetadataAbsent,
                          objects.SnapshotList.get_all,
                          self.context, search_opts)

    @mock.patch('cinder.objects.volume.Volume.get_by_id')
    @mock.patch('cinder.db.snapshot_get_all')
    def test_get_all_with_metadata(self, snapshot_get_all, volume_get_by_id):
        fake_volume_obj = fake_volume.fake_volume_obj(self.context)
        volume_get_by_id.return_value = fake_volume_obj

        db_snapshot = copy.deepcopy(fake_db_snapshot)
        db_snapshot['snapshot_metadata'] = [{'key': 'fake_key',
                                             'value': 'fake_value'}]
        snapshot_get_all.return_value = [db_snapshot]

        search_opts = mock.sentinel.search_opts
        snapshots = objects.SnapshotList.get_all(
            self.context, search_opts)
        self.assertEqual(1, len(snapshots))

        snapshot_obj = copy.deepcopy(fake_snapshot_obj)
        snapshot_obj['metadata'] = {'fake_key': 'fake_value'}
        TestSnapshot._compare(self, snapshot_obj, snapshots[0])
        snapshot_get_all.assert_called_once_with(self.context, search_opts,
                                                 None, None, None, None, None)



# Copyright (c) 2016 Huawei Technologies Co., Ltd.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import json

from oslo_log import log as logging
from oslo_utils import excutils

from cinder import exception
from cinder.i18n import _
from cinder.volume.drivers.huawei import constants
from cinder.volume.drivers.huawei import huawei_utils

LOG = logging.getLogger(__name__)


class AbsReplicaOp(object):
    def __init__(self, client):
        self.client = client

    def create(self, **kwargs):
        pass

    def delete(self, replica_id):
        pass

    def protect_second(self, replica_id):
        pass

    def unprotect_second(self, replica_id):
        pass

    def sync(self, replica_id):
        pass

    def split(self, replica_id):
        pass

    def switch(self, replica_id):
        pass

    def is_primary(self, replica_info):
        flag = replica_info.get('ISPRIMARY')
        if flag and flag.lower() == 'true':
            return True
        return False

    def get_replica_info(self, replica_id):
        return {}

    def _is_status(self, status_key, status, replica_info):
        if type(status) in (list, tuple):
            return replica_info.get(status_key, '') in status
        if type(status) is str:
            return replica_info.get(status_key, '') == status

        return False

    def is_running_status(self, status, replica_info):
        return self._is_status(constants.REPLICA_RUNNING_STATUS_KEY,
                               status, replica_info)

    def is_health_status(self, status, replica_info):
        return self._is_status(constants.REPLICA_HEALTH_STATUS_KEY,
                               status, replica_info)


class PairOp(AbsReplicaOp):
    def create(self, local_lun_id, rmt_lun_id, rmt_dev_id,
               rmt_dev_name, replica_model,
               speed=constants.REPLICA_SPEED,
               period=constants.REPLICA_PERIOD,
               **kwargs):
        super(PairOp, self).create(**kwargs)

        params = {
            "LOCALRESID": local_lun_id,
            "LOCALRESTYPE": '11',
            "REMOTEDEVICEID": rmt_dev_id,
            "REMOTEDEVICENAME": rmt_dev_name,
            "REMOTERESID": rmt_lun_id,
            "REPLICATIONMODEL": replica_model,
            # recovery policy. 1: auto, 2: manual
            "RECOVERYPOLICY": '1',
            "SPEED": speed,
        }

        if replica_model == constants.REPLICA_ASYNC_MODEL:
            # Synchronize type values:
            # 1, manual
            # 2, timed wait when synchronization begins
            # 3, timed wait when synchronization ends
            params['SYNCHRONIZETYPE'] = '2'
            params['TIMINGVAL'] = period

        try:
            pair_info = self.client.create_pair(params)
        except Exception as err:
            msg = _('Create replication pair failed. Error: %s.') % err
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        return pair_info

    def split(self, pair_id):
        self.client.split_pair(pair_id)

    def delete(self, pair_id, force=False):
        self.client.delete_pair(pair_id, force)

    def protect_second(self, pair_id):
        self.client.set_pair_second_access(pair_id,
                                           constants.REPLICA_SECOND_RO)

    def unprotect_second(self, pair_id):
        self.client.set_pair_second_access(pair_id,
                                           constants.REPLICA_SECOND_RW)

    def sync(self, pair_id):
        self.client.sync_pair(pair_id)

    def switch(self, pair_id):
        self.client.switch_pair(pair_id)

    def get_replica_info(self, pair_id):
        return self.client.get_pair_by_id(pair_id)


class CGOp(AbsReplicaOp):
    pass


class ReplicaCommonDriver(object):
    def __init__(self, conf, replica_op):
        self.conf = conf
        self.op = replica_op

    def protect_second(self, replica_id):
        info = self.op.get_replica_info(replica_id)
        if info.get('SECRESACCESS') == constants.REPLICA_SECOND_RO:
            return

        self.op.protect_second(replica_id)
        self.wait_second_access(replica_id, constants.REPLICA_SECOND_RO)

    def unprotect_second(self, replica_id):
        info = self.op.get_replica_info(replica_id)
        if info.get('SECRESACCESS') == constants.REPLICA_SECOND_RW:
            return

        self.op.unprotect_second(replica_id)
        self.wait_second_access(replica_id, constants.REPLICA_SECOND_RW)

    def sync(self, replica_id, wait_complete=False):
        self.protect_second(replica_id)

        expect_status = (constants.REPLICA_RUNNING_STATUS_NORMAL,
                         constants.REPLICA_RUNNING_STATUS_SYNC,
                         constants.REPLICA_RUNNING_STATUS_INITIAL_SYNC)
        info = self.op.get_replica_info(replica_id)

        # When running status is synchronizing or normal,
        # it's not necessary to do synchronize again.
        if (info.get('REPLICATIONMODEL') == constants.REPLICA_SYNC_MODEL
                and self.op.is_running_status(expect_status, info)):
            return

        self.op.sync(replica_id)
        self.wait_expect_state(replica_id, expect_status)

        if wait_complete:
            self.wait_replica_ready(replica_id)

    def split(self, replica_id):
        running_status = (constants.REPLICA_RUNNING_STATUS_SPLIT,
                          constants.REPLICA_RUNNING_STATUS_INVALID,
                          constants.REPLICA_RUNNING_STATUS_ERRUPTED)
        info = self.op.get_replica_info(replica_id)
        if self.op.is_running_status(running_status, info):
            return

        try:
            self.op.split(replica_id)
        except Exception as err:
            LOG.warning('Split replication exception: %s.', err)

        try:
            self.wait_expect_state(replica_id, running_status)
        except Exception:
            msg = _('Split replication failed.')
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

    def enable(self, replica_id, wait_sync_complete=False):
        info = self.op.get_replica_info(replica_id)
        if not self.op.is_primary(info):
            self.switch(replica_id)
        self.sync(replica_id)
        return None

    def switch(self, replica_id):
        self.split(replica_id)
        self.unprotect_second(replica_id)
        self.op.switch(replica_id)

        # Wait to be primary
        def _wait_switch_to_primary():
            info = self.op.get_replica_info(replica_id)
            if self.op.is_primary(info):
                return True
            return False

        interval = constants.DEFAULT_REPLICA_WAIT_INTERVAL
        timeout = constants.DEFAULT_REPLICA_WAIT_TIMEOUT
        huawei_utils.wait_for_condition(_wait_switch_to_primary,
                                        interval,
                                        timeout)

    def failover(self, replica_id):
        """Failover replication.

        Purpose:
            1. Split replication.
            2. Set secondary access read & write.
        """
        info = self.op.get_replica_info(replica_id)
        if self.op.is_primary(info):
            msg = _('We should not do switch over on primary array.')
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        sync_status_set = (constants.REPLICA_RUNNING_STATUS_SYNC,
                           constants.REPLICA_RUNNING_STATUS_INITIAL_SYNC)
        if self.op.is_running_status(sync_status_set, info):
            self.wait_replica_ready(replica_id)

        self.split(replica_id)
        self.op.unprotect_second(replica_id)

    def wait_replica_ready(self, replica_id, interval=None, timeout=None):
        LOG.debug('Wait synchronize complete.')
        running_status_normal = (constants.REPLICA_RUNNING_STATUS_NORMAL,
                                 constants.REPLICA_RUNNING_STATUS_SYNCED)
        running_status_sync = (constants.REPLICA_RUNNING_STATUS_SYNC,
                               constants.REPLICA_RUNNING_STATUS_INITIAL_SYNC)
        health_status_normal = constants.REPLICA_HEALTH_STATUS_NORMAL

        def _replica_ready():
            info = self.op.get_replica_info(replica_id)
            if (self.op.is_running_status(running_status_normal, info)
                    and self.op.is_health_status(health_status_normal, info)):
                return True

            if not self.op.is_running_status(running_status_sync, info):
                msg = (_('Wait synchronize failed. Running status: %s.') %
                       info.get(constants.REPLICA_RUNNING_STATUS_KEY))
                LOG.error(msg)
                raise exception.VolumeBackendAPIException(data=msg)

            return False

        if not interval:
            interval = constants.DEFAULT_WAIT_INTERVAL
        if not timeout:
            timeout = constants.DEFAULT_WAIT_TIMEOUT

        huawei_utils.wait_for_condition(_replica_ready,
                                        interval,
                                        timeout)

    def wait_second_access(self, replica_id, access_level):
        def _check_access():
            info = self.op.get_replica_info(replica_id)
            if info.get('SECRESACCESS') == access_level:
                return True
            return False

        interval = constants.DEFAULT_REPLICA_WAIT_INTERVAL
        timeout = constants.DEFAULT_REPLICA_WAIT_TIMEOUT
        huawei_utils.wait_for_condition(_check_access,
                                        interval,
                                        timeout)

    def wait_expect_state(self, replica_id,
                          running_status, health_status=None,
                          interval=None, timeout=None):
        def _check_state():
            info = self.op.get_replica_info(replica_id)
            if self.op.is_running_status(running_status, info):
                if (not health_status
                        or self.op.is_health_status(health_status, info)):
                    return True
            return False

        if not interval:
            interval = constants.DEFAULT_REPLICA_WAIT_INTERVAL
        if not timeout:
            timeout = constants.DEFAULT_REPLICA_WAIT_TIMEOUT

        huawei_utils.wait_for_condition(_check_state, interval, timeout)


def get_replication_driver_data(volume):
    if volume.replication_driver_data:
        return json.loads(volume.replication_driver_data)
    return {}


def to_string(dict_data):
    if dict_data:
        return json.dumps(dict_data)
    return ''


class ReplicaPairManager(object):
    def __init__(self, local_client, rmt_client, conf):
        self.local_client = local_client
        self.rmt_client = rmt_client
        self.conf = conf

        # Now just support one remote pool.
        self.rmt_pool = self.rmt_client.storage_pools[0]

        self.local_op = PairOp(self.local_client)
        self.local_driver = ReplicaCommonDriver(self.conf, self.local_op)
        self.rmt_op = PairOp(self.rmt_client)
        self.rmt_driver = ReplicaCommonDriver(self.conf, self.rmt_op)

    def try_get_remote_wwn(self):
        try:
            info = self.rmt_client.get_array_info()
            return info.get('wwn')
        except Exception as err:
            LOG.warning('Get remote array wwn failed. Error: %s.', err)
            return None

    def get_remote_device_by_wwn(self, wwn):
        devices = {}
        try:
            devices = self.local_client.get_remote_devices()
        except Exception as err:
            LOG.warning('Get remote devices failed. Error: %s.', err)

        for device in devices:
            if device.get('WWN') == wwn:
                return device

        return {}

    def check_remote_available(self):
        # We get device wwn in every check time.
        # If remote array changed, we can run normally.
        wwn = self.try_get_remote_wwn()
        if not wwn:
            return False

        device = self.get_remote_device_by_wwn(wwn)
        # Check remote device is available to use.
        # If array type is replication, 'ARRAYTYPE' == '1'.
        # If health status is normal, 'HEALTHSTATUS' == '1'.
        if (device and device.get('ARRAYTYPE') == '1'
                and device.get('HEALTHSTATUS') == '1'
                and device.get('RUNNINGSTATUS') == constants.STATUS_RUNNING):
            return True

        return False

    def update_replica_capability(self, stats):
        is_rmt_dev_available = self.check_remote_available()
        if not is_rmt_dev_available:
            LOG.warning('Remote device is unavailable.')
            return stats

        for pool in stats['pools']:
            pool['replication_enabled'] = True
            pool['replication_type'] = ['sync', 'async']

        return stats

    def get_rmt_dev_info(self):
        wwn = self.try_get_remote_wwn()
        if not wwn:
            return None, None

        device = self.get_remote_device_by_wwn(wwn)
        if not device:
            return None, None

        return device.get('ID'), device.get('NAME')

    def build_rmt_lun_params(self, local_lun_info):
        params = {
            'TYPE': '11',
            'NAME': local_lun_info['NAME'],
            'PARENTTYPE': '216',
            'PARENTID': self.rmt_client.get_pool_id(self.rmt_pool),
            'DESCRIPTION': local_lun_info['DESCRIPTION'],
            'ALLOCTYPE': local_lun_info['ALLOCTYPE'],
            'CAPACITY': local_lun_info['CAPACITY'],
            'READCACHEPOLICY': self.conf.lun_read_cache_policy,
            'WRITECACHEPOLICY': self.conf.lun_write_cache_policy,
        }

        if 'WRITEPOLICY' in local_lun_info:
            params['WRITEPOLICY'] = local_lun_info['WRITEPOLICY']
        if 'PREFETCHPOLICY' in local_lun_info:
            params['PREFETCHPOLICY'] = local_lun_info['PREFETCHPOLICY']
        if 'PREFETCHVALUE' in local_lun_info:
            params['PREFETCHVALUE'] = local_lun_info['PREFETCHVALUE']
        if 'DATATRANSFERPOLICY' in local_lun_info:
            params['DATATRANSFERPOLICY'] = local_lun_info['DATATRANSFERPOLICY']

        LOG.debug('Remote lun params: %s.', params)
        return params

    def wait_volume_online(self, client, lun_info,
                           interval=None, timeout=None):
        online_status = constants.STATUS_VOLUME_READY
        if lun_info.get('RUNNINGSTATUS') == online_status:
            return

        lun_id = lun_info['ID']

        def _wait_online():
            info = client.get_lun_info(lun_id)
            return info.get('RUNNINGSTATUS') == online_status

        if not interval:
            interval = constants.DEFAULT_REPLICA_WAIT_INTERVAL
        if not timeout:
            timeout = constants.DEFAULT_REPLICA_WAIT_TIMEOUT

        huawei_utils.wait_for_condition(_wait_online,
                                        interval,
                                        timeout)

    def create_rmt_lun(self, local_lun_info):
        # Create on rmt array. If failed, raise exception.
        lun_params = self.build_rmt_lun_params(local_lun_info)
        lun_info = self.rmt_client.create_lun(lun_params)
        try:
            self.wait_volume_online(self.rmt_client, lun_info)
        except exception.VolumeBackendAPIException:
            with excutils.save_and_reraise_exception():
                self.rmt_client.delete_lun(lun_info['ID'])

        return lun_info

    def create_replica(self, local_lun_info, replica_model):
        """Create remote LUN and replication pair.

        Purpose:
            1. create remote lun
            2. create replication pair
            3. enable replication pair
        """
        LOG.debug(('Create replication, local lun info: %(info)s, '
                   'replication model: %(model)s.'),
                  {'info': local_lun_info, 'model': replica_model})

        local_lun_id = local_lun_info['ID']
        self.wait_volume_online(self.local_client, local_lun_info)

        # step1, create remote lun
        rmt_lun_info = self.create_rmt_lun(local_lun_info)
        rmt_lun_id = rmt_lun_info['ID']

        # step2, get remote device info
        rmt_dev_id, rmt_dev_name = self.get_rmt_dev_info()
        if not rmt_lun_id or not rmt_dev_name:
            self._delete_rmt_lun(rmt_lun_id)
            msg = _('Get remote device info failed.')
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        # step3, create replication pair
        try:
            pair_info = self.local_op.create(local_lun_id,
                                             rmt_lun_id, rmt_dev_id,
                                             rmt_dev_name, replica_model)
            pair_id = pair_info['ID']
        except Exception as err:
            with excutils.save_and_reraise_exception():
                LOG.error('Create pair failed. Error: %s.', err)
                self._delete_rmt_lun(rmt_lun_id)

        # step4, start sync manually. If replication type is sync,
        # then wait for sync complete.
        wait_complete = (replica_model == constants.REPLICA_SYNC_MODEL)
        try:
            self.local_driver.sync(pair_id, wait_complete)
        except Exception as err:
            with excutils.save_and_reraise_exception():
                LOG.error('Start synchronization failed. Error: %s.', err)
                self._delete_pair(pair_id)
                self._delete_rmt_lun(rmt_lun_id)

        model_update = {}
        driver_data = {'pair_id': pair_id,
                       'rmt_lun_id': rmt_lun_id,
                       'rmt_lun_wwn': rmt_lun_info['WWN']}
        model_update['replication_driver_data'] = to_string(driver_data)
        model_update['replication_status'] = 'available'
        LOG.debug('Create replication, return info: %s.', model_update)
        return model_update

    def _delete_pair(self, pair_id):
        if (not pair_id
                or not self.local_client.check_pair_exist(pair_id)):
            return

        self.local_driver.split(pair_id)
        self.local_op.delete(pair_id)

    def _delete_rmt_lun(self, lun_id):
        if lun_id and self.rmt_client.check_lun_exist(lun_id):
            self.rmt_client.delete_lun(lun_id)

    def delete_replica(self, volume):
        """Delete replication pair and remote lun.

        Purpose:
            1. delete replication pair
            2. delete remote_lun
        """
        LOG.debug('Delete replication, volume: %s.', volume.id)
        info = get_replication_driver_data(volume)
        pair_id = info.get('pair_id')
        if pair_id:
            self._delete_pair(pair_id)

        # Delete remote_lun
        rmt_lun_id = info.get('rmt_lun_id')
        if rmt_lun_id:
            self._delete_rmt_lun(rmt_lun_id)

    def failback(self, volumes):
        """Failover volumes back to primary backend.

        The main steps:
        1. Switch the role of replication pairs.
        2. Copy the second LUN data back to primary LUN.
        3. Split replication pairs.
        4. Switch the role of replication pairs.
        5. Enable replications.
        """
        volumes_update = []
        for v in volumes:
            v_update = {}
            v_update['volume_id'] = v.id
            drv_data = get_replication_driver_data(v)
            pair_id = drv_data.get('pair_id')
            if not pair_id:
                LOG.warning("No pair id in volume %s.", v.id)
                v_update['updates'] = {'replication_status': 'error'}
                volumes_update.append(v_update)
                continue

            rmt_lun_id = drv_data.get('rmt_lun_id')
            if not rmt_lun_id:
                LOG.warning("No remote lun id in volume %s.", v.id)
                v_update['updates'] = {'replication_status': 'error'}
                volumes_update.append(v_update)
                continue

            # Switch replication pair role, and start synchronize.
            self.local_driver.enable(pair_id)

            # Wait for synchronize complete.
            self.local_driver.wait_replica_ready(pair_id)

            # Split replication pair again
            self.rmt_driver.failover(pair_id)

            # Switch replication pair role, and start synchronize.
            self.rmt_driver.enable(pair_id)

            local_metadata = huawei_utils.get_volume_private_data(v)
            new_drv_data = to_string(
                {'pair_id': pair_id,
                 'rmt_lun_id': local_metadata.get('huawei_lun_id'),
                 'rmt_lun_wwn': local_metadata.get('huawei_lun_wwn')})
            location = huawei_utils.to_string(
                huawei_lun_id=rmt_lun_id,
                huawei_lun_wwn=drv_data.get('rmt_lun_wwn'))

            v_update['updates'] = {'provider_location': location,
                                   'replication_status': 'available',
                                   'replication_driver_data': new_drv_data}
            volumes_update.append(v_update)

        return volumes_update

    def failover(self, volumes):
        """Failover volumes back to secondary array.

        Split the replication pairs and make the secondary LUNs R&W.
        """
        volumes_update = []
        for v in volumes:
            v_update = {}
            v_update['volume_id'] = v.id
            drv_data = get_replication_driver_data(v)
            pair_id = drv_data.get('pair_id')
            if not pair_id:
                LOG.warning("No pair id in volume %s.", v.id)
                v_update['updates'] = {'replication_status': 'error'}
                volumes_update.append(v_update)
                continue

            rmt_lun_id = drv_data.get('rmt_lun_id')
            if not rmt_lun_id:
                LOG.warning("No remote lun id in volume %s.", v.id)
                v_update['updates'] = {'replication_status': 'error'}
                volumes_update.append(v_update)
                continue

            self.rmt_driver.failover(pair_id)

            local_metadata = huawei_utils.get_volume_private_data(v)
            new_drv_data = to_string(
                {'pair_id': pair_id,
                 'rmt_lun_id': local_metadata.get('huawei_lun_id'),
                 'rmt_lun_wwn': local_metadata.get('huawei_lun_wwn')})
            location = huawei_utils.to_string(
                huawei_lun_id=rmt_lun_id,
                huawei_lun_wwn=drv_data.get('rmt_lun_wwn'))

            v_update['updates'] = {'provider_location': location,
                                   'replication_status': 'failed-over',
                                   'replication_driver_data': new_drv_data}
            volumes_update.append(v_update)

        return volumes_update


def get_replication_opts(opts):
    if opts.get('replication_type') == 'sync':
        opts['replication_type'] = constants.REPLICA_SYNC_MODEL
    else:
        opts['replication_type'] = constants.REPLICA_ASYNC_MODEL

    return opts



# Copyright (c) 2017 VMware, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Test suite for VMware vCenter FCD driver."""

from unittest import mock

import ddt
from oslo_utils import timeutils
from oslo_utils import units
from oslo_vmware import image_transfer
from oslo_vmware.objects import datastore
from oslo_vmware import vim_util

from cinder import context
from cinder import exception as cinder_exceptions
from cinder.tests.unit import fake_constants as fake
from cinder.tests.unit import fake_snapshot
from cinder.tests.unit import fake_volume
from cinder.tests.unit import test
from cinder.tests.unit import utils as test_utils
from cinder.volume import configuration
from cinder.volume.drivers.vmware import datastore as hub
from cinder.volume.drivers.vmware import fcd
from cinder.volume.drivers.vmware import vmdk
from cinder.volume.drivers.vmware import volumeops


@ddt.ddt
class VMwareVStorageObjectDriverTestCase(test.TestCase):

    IP = 'localhost'
    PORT = 2321
    IMG_TX_TIMEOUT = 10
    RESERVED_PERCENTAGE = 0
    VMDK_DRIVER = vmdk.VMwareVcVmdkDriver
    FCD_DRIVER = fcd.VMwareVStorageObjectDriver
    VC_VERSION = "6.7.0"

    VOL_ID = 'abcdefab-cdef-abcd-efab-cdefabcdefab'
    SRC_VOL_ID = '9b3f6f1b-03a9-4f1e-aaff-ae15122b6ccf'
    DISPLAY_NAME = 'foo'
    VOL_TYPE_ID = 'd61b8cb3-aa1b-4c9b-b79e-abcdbda8b58a'
    VOL_SIZE = 2
    PROJECT_ID = 'd45beabe-f5de-47b7-b462-0d9ea02889bc'
    IMAGE_ID = 'eb87f4b0-d625-47f8-bb45-71c43b486d3a'
    IMAGE_NAME = 'image-1'

    def setUp(self):
        super(VMwareVStorageObjectDriverTestCase, self).setUp()

        self._config = mock.Mock(spec=configuration.Configuration)
        self._config.vmware_host_ip = self.IP
        self._config.vmware_host_port = self.PORT
        self._config.vmware_image_transfer_timeout_secs = self.IMG_TX_TIMEOUT
        self._config.reserved_percentage = self.RESERVED_PERCENTAGE
        self._driver = fcd.VMwareVStorageObjectDriver(
            configuration=self._config)
        self._driver._vc_version = self.VC_VERSION
        self._driver._storage_policy_enabled = True
        self._context = context.get_admin_context()
        self.updated_at = timeutils.utcnow()

    @mock.patch.object(VMDK_DRIVER, 'do_setup')
    @mock.patch.object(FCD_DRIVER, 'volumeops')
    def test_do_setup(self, vops, vmdk_do_setup):
        self._driver._storage_policy_enabled = False
        self._driver.do_setup(self._context)

        vmdk_do_setup.assert_called_once_with(self._context)
        vops.set_vmx_version.assert_called_once_with('vmx-13')
        self.assertTrue(self._driver._use_fcd_snapshot)
        self.assertTrue(self._driver._storage_policy_enabled)

    @mock.patch.object(VMDK_DRIVER, 'volumeops')
    @mock.patch.object(VMDK_DRIVER, '_get_datastore_summaries')
    def test_get_volume_stats(self, _get_datastore_summaries, vops):
        FREE_GB = 7
        TOTAL_GB = 11

        class ObjMock(object):
            def __init__(self, **kwargs):
                self.__dict__.update(kwargs)

        _get_datastore_summaries.return_value = \
            ObjMock(objects= [
                ObjMock(propSet = [
                    ObjMock(name = "host",
                            val = ObjMock(DatastoreHostMount = [])),
                    ObjMock(name = "summary",
                            val = ObjMock(freeSpace = FREE_GB * units.Gi,
                                          capacity = TOTAL_GB * units.Gi,
                                          accessible = True))
                ])
            ])

        vops._in_maintenance.return_value = False

        stats = self._driver.get_volume_stats()

        self.assertEqual('VMware', stats['vendor_name'])
        self.assertEqual(self._driver.VERSION, stats['driver_version'])
        self.assertEqual(self._driver.STORAGE_TYPE, stats['storage_protocol'])
        self.assertEqual(self.RESERVED_PERCENTAGE,
                         stats['reserved_percentage'])
        self.assertEqual(TOTAL_GB, stats['total_capacity_gb'])
        self.assertEqual(FREE_GB, stats['free_capacity_gb'])

    def _create_volume_dict(self,
                            vol_id=VOL_ID,
                            display_name=DISPLAY_NAME,
                            volume_type_id=VOL_TYPE_ID,
                            status='available',
                            size=VOL_SIZE,
                            attachment=None,
                            project_id=PROJECT_ID):
        return {'id': vol_id,
                'display_name': display_name,
                'name': 'volume-%s' % vol_id,
                'volume_type_id': volume_type_id,
                'status': status,
                'size': size,
                'volume_attachment': attachment,
                'project_id': project_id,
                }

    def _create_volume_obj(self,
                           vol_id=VOL_ID,
                           display_name=DISPLAY_NAME,
                           volume_type_id=VOL_TYPE_ID,
                           status='available',
                           size=VOL_SIZE,
                           attachment=None,
                           project_id=PROJECT_ID):
        vol = self._create_volume_dict(
            vol_id, display_name, volume_type_id, status, size, attachment,
            project_id)
        return fake_volume.fake_volume_obj(self._context, **vol)

    @mock.patch.object(FCD_DRIVER, '_get_storage_profile')
    @mock.patch.object(FCD_DRIVER, '_select_datastore')
    def test_select_ds_fcd(self, select_datastore, get_storage_profile):
        profile = mock.sentinel.profile
        get_storage_profile.return_value = profile

        datastore = mock.sentinel.datastore
        summary = mock.Mock(datastore=datastore)
        select_datastore.return_value = (mock.ANY, mock.ANY, summary)

        volume = self._create_volume_obj()
        ret = self._driver._select_ds_fcd(volume)
        self.assertEqual(datastore, ret)
        exp_req = {hub.DatastoreSelector.SIZE_BYTES: volume.size * units.Gi,
                   hub.DatastoreSelector.PROFILE_NAME: profile}
        select_datastore.assert_called_once_with(exp_req)

    @mock.patch.object(FCD_DRIVER, '_select_datastore')
    @mock.patch.object(FCD_DRIVER, 'volumeops')
    def _test_get_temp_image_folder(
            self, vops, select_datastore, preallocated=False):
        host = mock.sentinel.host
        summary = mock.Mock()
        summary.name = 'ds-1'
        select_datastore.return_value = (host, mock.ANY, summary)

        dc_ref = mock.sentinel.dc_ref
        vops.get_dc.return_value = dc_ref

        size_bytes = units.Gi
        ret = self._driver._get_temp_image_folder(size_bytes, preallocated)
        self.assertEqual(
            (dc_ref, summary, vmdk.TMP_IMAGES_DATASTORE_FOLDER_PATH), ret)
        exp_req = {hub.DatastoreSelector.SIZE_BYTES: size_bytes}
        if preallocated:
            exp_req[hub.DatastoreSelector.HARD_AFFINITY_DS_TYPE] = (
                {hub.DatastoreType.NFS,
                 hub.DatastoreType.VMFS,
                 hub.DatastoreType.NFS41})
        select_datastore.assert_called_once_with(exp_req)
        vops.get_dc.assert_called_once_with(host)
        vops.create_datastore_folder.assert_called_once_with(
            summary.name, vmdk.TMP_IMAGES_DATASTORE_FOLDER_PATH, dc_ref)

    def test_get_temp_image_folder(self):
        self._test_get_temp_image_folder()

    def test_get_temp_image_folder_preallocated(self):
        self._test_get_temp_image_folder(preallocated=True)

    @mock.patch.object(VMDK_DRIVER, '_get_disk_type')
    @ddt.data(('eagerZeroedThick', 'eagerZeroedThick'),
              ('thick', 'preallocated'),
              ('thin', 'thin'))
    @ddt.unpack
    def test_get_disk_type(
            self, extra_spec_disk_type, exp_ret_val, vmdk_get_disk_type):
        vmdk_get_disk_type.return_value = extra_spec_disk_type

        volume = mock.sentinel.volume
        ret = self._driver._get_disk_type(volume)
        self.assertEqual(exp_ret_val, ret)

    @mock.patch.object(FCD_DRIVER, '_select_ds_fcd')
    @mock.patch.object(FCD_DRIVER, '_get_disk_type')
    @mock.patch.object(FCD_DRIVER, '_get_storage_profile_id')
    @mock.patch.object(FCD_DRIVER, 'volumeops')
    def test_create_volume(self, vops, get_storage_profile_id, get_disk_type,
                           select_ds_fcd):
        ds_ref = mock.sentinel.ds_ref
        select_ds_fcd.return_value = ds_ref

        disk_type = mock.sentinel.disk_type
        get_disk_type.return_value = disk_type

        profile_id = mock.sentinel.profile_id
        get_storage_profile_id.return_value = profile_id

        fcd_loc = mock.Mock()
        provider_loc = mock.sentinel.provider_loc
        fcd_loc.provider_location.return_value = provider_loc
        vops.create_fcd.return_value = fcd_loc

        volume = self._create_volume_obj()
        ret = self._driver.create_volume(volume)
        self.assertEqual({'provider_location': provider_loc}, ret)
        select_ds_fcd.assert_called_once_with(volume)
        get_disk_type.assert_called_once_with(volume)
        vops.create_fcd.assert_called_once_with(
            volume.name, volume.size * units.Ki, ds_ref, disk_type,
            profile_id=profile_id)

    @mock.patch.object(volumeops.FcdLocation, 'from_provider_location')
    @mock.patch.object(FCD_DRIVER, 'volumeops')
    def test_delete_fcd(self, vops, from_provider_loc):
        fcd_loc = mock.sentinel.fcd_loc
        from_provider_loc.return_value = fcd_loc

        provider_loc = mock.sentinel.provider_loc
        self._driver._delete_fcd(provider_loc)
        from_provider_loc.test_assert_called_once_with(provider_loc)
        vops.delete_fcd.assert_called_once_with(fcd_loc)

    @mock.patch.object(FCD_DRIVER, '_delete_fcd')
    def test_delete_volume(self, delete_fcd):
        volume = self._create_volume_obj()
        volume.provider_location = 'foo@ds1'
        self._driver.delete_volume(volume)
        delete_fcd.assert_called_once_with(volume.provider_location)

    @mock.patch.object(FCD_DRIVER, '_delete_fcd')
    def test_delete_volume_empty_provider_location(self, delete_fcd):
        volume = self._create_volume_obj()
        self._driver.delete_volume(volume)
        delete_fcd.assert_not_called()

    @mock.patch.object(volumeops.FcdLocation, 'from_provider_location')
    @mock.patch.object(FCD_DRIVER, '_get_adapter_type')
    def test_initialize_connection(
            self, get_adapter_type, from_provider_location):
        fcd_loc = mock.Mock(
            fcd_id=mock.sentinel.fcd_id, ds_ref_val=mock.sentinel.ds_ref_val)
        from_provider_location.return_value = fcd_loc

        adapter_type = mock.sentinel.adapter_type
        get_adapter_type.return_value = adapter_type

        volume = self._create_volume_obj()
        connector = mock.sentinel.connector
        ret = self._driver.initialize_connection(volume, connector)
        self.assertEqual(self._driver.STORAGE_TYPE, ret['driver_volume_type'])
        self.assertEqual(fcd_loc.fcd_id, ret['data']['id'])
        self.assertEqual(fcd_loc.ds_ref_val, ret['data']['ds_ref_val'])
        self.assertEqual(adapter_type, ret['data']['adapter_type'])

    def test_container_format(self):
        self._driver._validate_container_format('bare', mock.sentinel.image_id)

    def test_container_format_invalid(self):
        self.assertRaises(cinder_exceptions.ImageUnacceptable,
                          self._driver._validate_container_format,
                          'ova',
                          mock.sentinel.image_id)

    def _create_image_meta(self,
                           _id=IMAGE_ID,
                           name=IMAGE_NAME,
                           disk_format='vmdk',
                           size=1 * units.Gi,
                           container_format='bare',
                           vmware_disktype='streamOptimized',
                           vmware_adaptertype='lsiLogic',
                           is_public=True):
        return {'id': _id,
                'name': name,
                'disk_format': disk_format,
                'size': size,
                'container_format': container_format,
                'properties': {'vmware_disktype': vmware_disktype,
                               'vmware_adaptertype': vmware_adaptertype,
                               },
                'is_public': is_public,
                }

    @mock.patch.object(FCD_DRIVER, '_get_temp_image_folder')
    @mock.patch.object(FCD_DRIVER, '_create_virtual_disk_from_sparse_image')
    @mock.patch.object(FCD_DRIVER,
                       '_create_virtual_disk_from_preallocated_image')
    @mock.patch.object(FCD_DRIVER, 'volumeops')
    @mock.patch.object(datastore, 'DatastoreURL')
    @mock.patch.object(FCD_DRIVER, '_get_storage_profile_id')
    @ddt.data(vmdk.ImageDiskType.PREALLOCATED, vmdk.ImageDiskType.SPARSE,
              vmdk.ImageDiskType.STREAM_OPTIMIZED)
    def test_copy_image_to_volume(self,
                                  disk_type,
                                  get_storage_profile_id,
                                  datastore_url_cls,
                                  vops,
                                  create_disk_from_preallocated_image,
                                  create_disk_from_sparse_image,
                                  get_temp_image_folder):
        image_meta = self._create_image_meta(vmware_disktype=disk_type)
        image_service = mock.Mock()
        image_service.show.return_value = image_meta

        dc_ref = mock.sentinel.dc_ref
        datastore = mock.sentinel.datastore
        summary = mock.Mock(datastore=datastore)
        summary.name = 'ds1'
        folder_path = mock.sentinel.folder_path
        get_temp_image_folder.return_value = (dc_ref, summary, folder_path)

        vmdk_path = mock.Mock()
        vmdk_path.get_descriptor_ds_file_path.return_value = (
            "[ds1] cinder_vol/foo.vmdk")
        if disk_type == vmdk.ImageDiskType.PREALLOCATED:
            create_disk_from_preallocated_image.return_value = vmdk_path
        else:
            create_disk_from_sparse_image.return_value = vmdk_path

        dc_path = '/test-dc'
        vops.get_inventory_path.return_value = dc_path

        ds_url = mock.sentinel.ds_url
        datastore_url_cls.return_value = ds_url

        profile_id = mock.sentinel.profile_id
        get_storage_profile_id.return_value = profile_id

        fcd_loc = mock.Mock()
        provider_location = mock.sentinel.provider_location
        fcd_loc.provider_location.return_value = provider_location
        vops.register_disk.return_value = fcd_loc

        volume = self._create_volume_obj()
        image_id = self.IMAGE_ID
        ret = self._driver.copy_image_to_volume(
            self._context, volume, image_service, image_id)

        self.assertEqual({'provider_location': provider_location}, ret)
        get_temp_image_folder.assert_called_once_with(volume.size * units.Gi)
        if disk_type == vmdk.ImageDiskType.PREALLOCATED:
            create_disk_from_preallocated_image.assert_called_once_with(
                self._context, image_service, image_id, image_meta['size'],
                dc_ref, summary.name, folder_path, volume.id,
                volumeops.VirtualDiskAdapterType.LSI_LOGIC)
        else:
            create_disk_from_sparse_image.assert_called_once_with(
                self._context, image_service, image_id, image_meta['size'],
                dc_ref, summary.name, folder_path, volume.id)
        datastore_url_cls.assert_called_once_with(
            'https', self._driver.configuration.vmware_host_ip,
            'cinder_vol/foo.vmdk', '/test-dc', 'ds1')
        vops.register_disk.assert_called_once_with(
            str(ds_url),
            volume.name,
            summary.datastore)
        vops.update_fcd_policy.assert_called_once_with(fcd_loc, profile_id)

    @mock.patch.object(volumeops.FcdLocation, 'from_provider_location')
    @mock.patch.object(FCD_DRIVER, 'volumeops')
    @mock.patch.object(vim_util, 'get_moref')
    @mock.patch.object(FCD_DRIVER, '_create_backing')
    @mock.patch.object(image_transfer, 'upload_image')
    @mock.patch.object(VMDK_DRIVER, 'session')
    @mock.patch.object(FCD_DRIVER, '_delete_temp_backing')
    def test_copy_volume_to_image(
            self, delete_temp_backing, session, upload_image, create_backing,
            get_moref, vops, from_provider_loc):
        fcd_loc = mock.Mock()
        ds_ref = mock.sentinel.ds_ref
        fcd_loc.ds_ref.return_value = ds_ref
        from_provider_loc.return_value = fcd_loc

        host_ref_val = mock.sentinel.host_ref_val
        vops.get_connected_hosts.return_value = [host_ref_val]

        host = mock.sentinel.host
        get_moref.return_value = host

        backing = mock.sentinel.backing
        create_backing.return_value = backing

        vmdk_file_path = mock.sentinel.vmdk_file_path
        vops.get_vmdk_path.return_value = vmdk_file_path
        vops.get_backing_by_uuid.return_value = backing

        volume = test_utils.create_volume(
            self._context, volume_type_id=fake.VOLUME_TYPE_ID,
            updated_at=self.updated_at)
        extra_specs = {
            'image_service:store_id': 'fake-store'
        }
        test_utils.create_volume_type(
            self._context.elevated(), id=fake.VOLUME_TYPE_ID,
            name="test_type", extra_specs=extra_specs)

        image_service = mock.sentinel.image_service
        image_meta = self._create_image_meta()
        self._driver.copy_volume_to_image(
            self._context, volume, image_service, image_meta)

        from_provider_loc.assert_called_once_with(volume.provider_location)
        vops.get_connected_hosts.assert_called_once_with(ds_ref)
        create_backing.assert_called_once_with(
            volume, host, {vmdk.CREATE_PARAM_DISK_LESS: True})
        vops.attach_fcd.assert_called_once_with(backing, fcd_loc)
        vops.get_vmdk_path.assert_called_once_with(backing)
        conf = self._driver.configuration
        upload_image.assert_called_once_with(
            self._context,
            conf.vmware_image_transfer_timeout_secs,
            image_service,
            image_meta['id'],
            volume.project_id,
            session=session,
            host=conf.vmware_host_ip,
            port=conf.vmware_host_port,
            vm=backing,
            vmdk_file_path=vmdk_file_path,
            vmdk_size=volume.size * units.Gi,
            image_name=image_meta['name'],
            store_id='fake-store',
            base_image_ref=None)
        vops.detach_fcd.assert_called_once_with(backing, fcd_loc)
        delete_temp_backing.assert_called_once_with(backing)

    @mock.patch.object(volumeops.FcdLocation, 'from_provider_location')
    @mock.patch.object(FCD_DRIVER, 'volumeops')
    def test_extend_volume(self, vops, from_provider_loc):
        fcd_loc = mock.sentinel.fcd_loc
        from_provider_loc.return_value = fcd_loc

        volume = self._create_volume_obj()
        new_size = 3
        self._driver.extend_volume(volume, new_size)
        from_provider_loc.assert_called_once_with(volume.provider_location)
        vops.extend_fcd.assert_called_once_with(
            fcd_loc, new_size * units.Ki)

    @mock.patch.object(volumeops.FcdLocation, 'from_provider_location')
    @mock.patch.object(FCD_DRIVER, 'volumeops')
    def test_clone_fcd(self, vops, from_provider_loc):
        fcd_loc = mock.sentinel.fcd_loc
        from_provider_loc.return_value = fcd_loc

        dest_fcd_loc = mock.sentinel.dest_fcd_loc
        vops.clone_fcd.return_value = dest_fcd_loc

        provider_loc = mock.sentinel.provider_loc
        name = mock.sentinel.name
        dest_ds_ref = mock.sentinel.dest_ds_ref
        disk_type = mock.sentinel.disk_type
        ret = self._driver._clone_fcd(
            provider_loc, name, dest_ds_ref, disk_type)
        self.assertEqual(dest_fcd_loc, ret)
        from_provider_loc.assert_called_once_with(provider_loc)
        vops.clone_fcd.assert_called_once_with(
            name, fcd_loc, dest_ds_ref, disk_type, profile_id=None)

    @mock.patch.object(FCD_DRIVER, '_select_ds_fcd')
    @mock.patch.object(FCD_DRIVER, '_clone_fcd')
    @mock.patch.object(FCD_DRIVER, 'volumeops')
    @mock.patch.object(volumeops.FcdLocation, 'from_provider_location')
    def _test_create_snapshot(
            self, from_provider_loc, vops, clone_fcd, select_ds_fcd,
            use_fcd_snapshot=False):
        self._driver._use_fcd_snapshot = use_fcd_snapshot

        provider_location = mock.sentinel.provider_location
        if use_fcd_snapshot:
            fcd_loc = mock.sentinel.fcd_loc
            from_provider_loc.return_value = fcd_loc

            fcd_snap_loc = mock.Mock()
            fcd_snap_loc.provider_location.return_value = provider_location
            vops.create_fcd_snapshot.return_value = fcd_snap_loc
        else:
            ds_ref = mock.sentinel.ds_ref
            select_ds_fcd.return_value = ds_ref

            dest_fcd_loc = mock.Mock()
            dest_fcd_loc.provider_location.return_value = provider_location
            clone_fcd.return_value = dest_fcd_loc

        volume = self._create_volume_obj()
        snapshot = fake_snapshot.fake_snapshot_obj(
            self._context, volume=volume)
        ret = self._driver.create_snapshot(snapshot)
        self.assertEqual({'provider_location': provider_location}, ret)

        if use_fcd_snapshot:
            vops.create_fcd_snapshot.assert_called_once_with(
                fcd_loc, description="snapshot-%s" % snapshot.id)
        else:
            select_ds_fcd.assert_called_once_with(snapshot.volume)
            clone_fcd.assert_called_once_with(
                volume.provider_location, snapshot.name, ds_ref)

    def test_create_snapshot_legacy(self):
        self._test_create_snapshot()

    def test_create_snapshot(self):
        self._test_create_snapshot(use_fcd_snapshot=True)

    @mock.patch.object(FCD_DRIVER, '_delete_fcd')
    @mock.patch.object(volumeops.FcdSnapshotLocation, 'from_provider_location')
    @mock.patch.object(FCD_DRIVER, 'volumeops')
    def _test_delete_snapshot(
            self, vops, from_provider_loc, delete_fcd,
            empty_provider_loc=False, use_fcd_snapshot=False):
        volume = self._create_volume_obj()
        snapshot = fake_snapshot.fake_snapshot_obj(
            self._context, volume=volume)

        if empty_provider_loc:
            snapshot.provider_location = None
        else:
            snapshot.provider_location = "test"
            if use_fcd_snapshot:
                fcd_snap_loc = mock.sentinel.fcd_snap_loc
                from_provider_loc.return_value = fcd_snap_loc
            else:
                from_provider_loc.return_value = None

        self._driver.delete_snapshot(snapshot)
        if empty_provider_loc:
            delete_fcd.assert_not_called()
            vops.delete_fcd_snapshot.assert_not_called()
        elif use_fcd_snapshot:
            vops.delete_fcd_snapshot.assert_called_once_with(fcd_snap_loc)
        else:
            delete_fcd.assert_called_once_with(snapshot.provider_location)

    def test_delete_snapshot_legacy(self):
        self._test_delete_snapshot()

    def test_delete_snapshot_with_empty_provider_loc(self):
        self._test_delete_snapshot(empty_provider_loc=True)

    def test_delete_snapshot(self):
        self._test_delete_snapshot(use_fcd_snapshot=True)

    @mock.patch.object(FCD_DRIVER, 'volumeops')
    @ddt.data((1, 1), (1, 2))
    @ddt.unpack
    def test_extend_if_needed(self, cur_size, new_size, vops):
        fcd_loc = mock.sentinel.fcd_loc
        self._driver._extend_if_needed(fcd_loc, cur_size, new_size)
        if new_size > cur_size:
            vops.extend_fcd.assert_called_once_with(
                fcd_loc, new_size * units.Ki)
        else:
            vops.extend_fcd.assert_not_called()

    @mock.patch.object(FCD_DRIVER, '_select_ds_fcd')
    @mock.patch.object(FCD_DRIVER, '_get_disk_type')
    @mock.patch.object(FCD_DRIVER, '_get_storage_profile_id')
    @mock.patch.object(FCD_DRIVER, '_clone_fcd')
    @mock.patch.object(FCD_DRIVER, '_extend_if_needed')
    def test_create_volume_from_fcd(
            self, extend_if_needed, clone_fcd, get_storage_profile_id,
            get_disk_type, select_ds_fcd):
        ds_ref = mock.sentinel.ds_ref
        select_ds_fcd.return_value = ds_ref

        disk_type = mock.sentinel.disk_type
        get_disk_type.return_value = disk_type

        profile_id = mock.sentinel.profile_id
        get_storage_profile_id.return_value = profile_id

        cloned_fcd_loc = mock.Mock()
        dest_provider_loc = mock.sentinel.dest_provider_loc
        cloned_fcd_loc.provider_location.return_value = dest_provider_loc
        clone_fcd.return_value = cloned_fcd_loc

        provider_loc = mock.sentinel.provider_loc
        cur_size = 1
        volume = self._create_volume_obj()
        ret = self._driver._create_volume_from_fcd(
            provider_loc, cur_size, volume)
        self.assertEqual({'provider_location': dest_provider_loc}, ret)
        select_ds_fcd.test_assert_called_once_with(volume)
        get_disk_type.test_assert_called_once_with(volume)
        clone_fcd.assert_called_once_with(
            provider_loc, volume.name, ds_ref, disk_type=disk_type,
            profile_id=profile_id)
        extend_if_needed.assert_called_once_with(
            cloned_fcd_loc, cur_size, volume.size)

    @mock.patch.object(FCD_DRIVER, '_create_volume_from_fcd')
    @mock.patch.object(volumeops.FcdSnapshotLocation, 'from_provider_location')
    @mock.patch.object(FCD_DRIVER, '_get_storage_profile_id')
    @mock.patch.object(FCD_DRIVER, 'volumeops')
    @mock.patch.object(FCD_DRIVER, '_extend_if_needed')
    def _test_create_volume_from_snapshot(
            self, extend_if_needed, vops, get_storage_profile_id,
            from_provider_loc, create_volume_from_fcd, use_fcd_snapshot=False):
        src_volume = self._create_volume_obj(vol_id=self.SRC_VOL_ID)
        snapshot = fake_snapshot.fake_snapshot_obj(
            self._context, volume=src_volume)
        volume = self._create_volume_obj(size=self.VOL_SIZE + 1)

        if use_fcd_snapshot:
            fcd_snap_loc = mock.sentinel.fcd_snap_loc
            from_provider_loc.return_value = fcd_snap_loc

            profile_id = mock.sentinel.profile_id
            get_storage_profile_id.return_value = profile_id

            fcd_loc = mock.Mock()
            provider_loc = mock.sentinel.provider_loc
            fcd_loc.provider_location.return_value = provider_loc
            vops.create_fcd_from_snapshot.return_value = fcd_loc
        else:
            from_provider_loc.return_value = None

        ret = self._driver.create_volume_from_snapshot(volume, snapshot)
        if use_fcd_snapshot:
            self.assertEqual({'provider_location': provider_loc}, ret)
            vops.create_fcd_from_snapshot.assert_called_once_with(
                fcd_snap_loc, volume.name, profile_id=profile_id)
            extend_if_needed.assert_called_once_with(
                fcd_loc, snapshot.volume_size, volume.size)
        else:
            create_volume_from_fcd.assert_called_once_with(
                snapshot.provider_location, snapshot.volume.size, volume)

    def test_create_volume_from_snapshot_legacy(self):
        self._test_create_volume_from_snapshot()

    def test_create_volume_from_snapshot(self):
        self._test_create_volume_from_snapshot(use_fcd_snapshot=True)

    @mock.patch.object(FCD_DRIVER, '_create_volume_from_fcd')
    def test_create_cloned_volume(self, create_volume_from_fcd):
        src_volume = self._create_volume_obj()
        volume = mock.sentinel.volume
        self._driver.create_cloned_volume(volume, src_volume)
        create_volume_from_fcd.assert_called_once_with(
            src_volume.provider_location, src_volume.size, volume)

    @mock.patch.object(FCD_DRIVER, '_get_storage_profile')
    @mock.patch.object(FCD_DRIVER, '_get_extra_spec_storage_profile')
    @mock.patch.object(FCD_DRIVER, '_in_use')
    @mock.patch.object(FCD_DRIVER, 'volumeops')
    @mock.patch.object(volumeops.FcdLocation, 'from_provider_location')
    @mock.patch.object(FCD_DRIVER, 'ds_sel')
    @ddt.data({},
              {'storage_policy_enabled': False},
              {'same_profile': True},
              {'vol_in_use': True}
              )
    @ddt.unpack
    def test_retype(
            self, ds_sel, from_provider_location, vops, in_use,
            get_extra_spec_storage_profile, get_storage_profile,
            storage_policy_enabled=True, same_profile=False, vol_in_use=False):
        self._driver._storage_policy_enabled = storage_policy_enabled

        if storage_policy_enabled:
            profile = mock.sentinel.profile
            get_storage_profile.return_value = profile

            if same_profile:
                new_profile = profile
            else:
                new_profile = mock.sentinel.new_profile
            get_extra_spec_storage_profile.return_value = new_profile

            in_use.return_value = vol_in_use

            if not vol_in_use:
                fcd_loc = mock.sentinel.fcd_loc
                from_provider_location.return_value = fcd_loc

                new_profile_id = mock.Mock()
                ds_sel.get_profile_id.return_value = new_profile_id

        ctxt = mock.sentinel.ctxt
        volume = self._create_volume_obj()
        new_type = {'id': mock.sentinel.new_type_id}
        diff = mock.sentinel.diff
        host = mock.sentinel.host
        ret = self._driver.retype(ctxt, volume, new_type, diff, host)

        if not storage_policy_enabled:
            self.assertTrue(ret)
        else:
            get_storage_profile.assert_called_once_with(volume)
            get_extra_spec_storage_profile.assert_called_once_with(
                new_type['id'])
            if same_profile:
                self.assertTrue(ret)
            else:
                in_use.assert_called_once_with(volume)
                if vol_in_use:
                    self.assertFalse(ret)
                else:
                    ds_sel.get_profile_id.assert_called_once_with(new_profile)
                    vops.update_fcd_policy.assert_called_once_with(
                        fcd_loc, new_profile_id.uniqueId)



# Copyright 2018 FiberHome Telecommunication Technologies CO.,LTD
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from oslo_log import log as logging
from oslo_utils import strutils
from six.moves import http_client
from webob import exc

from cinder.api import common
from cinder.api.contrib import volume_transfer as volume_transfer_v2
from cinder.api import microversions as mv
from cinder.api.openstack import wsgi
from cinder.api.schemas import volume_transfer
from cinder.api import validation
from cinder import exception

LOG = logging.getLogger(__name__)


class VolumeTransferController(volume_transfer_v2.VolumeTransferController):
    """The transfer API controller for the OpenStack API V3."""

    def _get_transfers(self, req, is_detail):
        """Returns a list of transfers, transformed through view builder."""
        context = req.environ['cinder.context']
        req_version = req.api_version_request
        params = req.params.copy()
        marker = limit = offset = None
        if req_version.matches(mv.SUPPORT_TRANSFER_PAGINATION):
            marker, limit, offset = common.get_pagination_params(params)
            sort_keys, sort_dirs = common.get_sort_params(params)
        else:
            # NOTE(yikun): After microversion SUPPORT_TRANSFER_PAGINATION,
            # transfers list api use the ['created_at'], ['asc']
            # as default order, but we should keep the compatible in here.
            sort_keys, sort_dirs = ['created_at', 'id'], ['asc', 'asc']
        filters = params
        if 'name' in filters:
            filters['display_name'] = filters.pop('name')
        LOG.debug('Listing volume transfers')

        transfers = self.transfer_api.get_all(context, marker=marker,
                                              limit=limit,
                                              sort_keys=sort_keys,
                                              sort_dirs=sort_dirs,
                                              filters=filters,
                                              offset=offset)
        transfer_count = len(transfers)
        limited_list = common.limited(transfers, req)

        if is_detail:
            transfers = self._view_builder.detail_list(req, limited_list,
                                                       transfer_count)
        else:
            transfers = self._view_builder.summary_list(req, limited_list,
                                                        transfer_count)

        return transfers

    def index(self, req):
        """Returns a summary list of transfers."""
        return self._get_transfers(req, is_detail=False)

    def detail(self, req):
        """Returns a detailed list of transfers."""
        return self._get_transfers(req, is_detail=True)

    @wsgi.response(http_client.ACCEPTED)
    @validation.schema(volume_transfer.create, mv.BASE_VERSION,
                       mv.get_prior_version(mv.TRANSFER_WITH_SNAPSHOTS))
    @validation.schema(volume_transfer.create_v355, mv.TRANSFER_WITH_SNAPSHOTS)
    def create(self, req, body):
        """Create a new volume transfer."""
        LOG.debug('Creating new volume transfer %s', body)

        context = req.environ['cinder.context']
        transfer = body['transfer']

        volume_id = transfer['volume_id']

        name = transfer.get('name', None)
        if name is not None:
            name = name.strip()

        no_snapshots = strutils.bool_from_string(transfer.get('no_snapshots',
                                                              False))

        LOG.info("Creating transfer of volume %s", volume_id)

        try:
            new_transfer = self.transfer_api.create(context, volume_id, name,
                                                    no_snapshots=no_snapshots)
        # Not found exception will be handled at the wsgi level
        except exception.Invalid as error:
            raise exc.HTTPBadRequest(explanation=error.msg)

        transfer = self._view_builder.create(req,
                                             dict(new_transfer))
        return transfer


def create_resource():
    return wsgi.Resource(VolumeTransferController())
